/*******************************************************************************
* (c) Copyright 2007 MySQL, Inc., All Rights Reserved.
* This software and documentation are copyright protected works that include
* confidential information and trade secrets proprietary to MySQL, Inc.
* and may not be copied, duplicated, disclosed or used, in whole
* or in part, except pursuant to the License Agreement or as otherwise expressly
* approved by MySQL, inc..
*
* @File mgmvm.cpp
*
* @Author   dkehn@mysql.com - MySQL, Inc.
* @date 18-JUL-2007
* @version: 1.0
*
* @brief This code is intended to present the MemoryUsage of the data nodes
*        in a mysql cluster.
* 
* @example:
*
* Revision History:  (latest changes at top)
*    don@mysql.com - 17-JUL-2007 - created
*
*/
#include <mysql.h>
#include <ndbapi/NdbApi.hpp>
#include <mgmapi.h>
#include <mgmapi_debug.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>

typedef struct _nodeStat {
    int nodeID;
    unsigned page_size;
    unsigned used;
    unsigned total;
    unsigned blocks;
    float percent_used;
 } nodeStat;
    
typedef struct _statInfo {
    int numNodes;
    nodeStat ns[64];
} statInfo;
        
typedef struct _nodeStats {
    int statCmdSent;
    int nodeid;
    int dumpSent;
    int dumpDnRecv;
    int dumpIdxRecv;
    int statRecv;
    int nodeState;

    char memdatabuf[80];
    char memidxbuf[80];
    char countbuf[80];
} NODESTATS;

#define LINES_BETWEEN_HEADER           20
#define MAX_DISPLAY_NODES              63
NODESTATS Nodes[MAX_DISPLAY_NODES];

int displayNodeCount = 0;

#define MGMERROR(h) \
{ \
  fprintf(stderr, "code: %d msg: %s\n", \
          ndb_mgm_get_latest_error(h), \
          ndb_mgm_get_latest_error_msg(h)); \
  exit(-1); \
}

#define LOGEVENTERROR(h) \
{ \
  fprintf(stderr, "code: %d msg: %s\n", \
          ndb_logevent_get_latest_error(h), \
          ndb_logevent_get_latest_error_msg(h)); \
  exit(-1); \
}
static char *prgName;
static char *version = "1.0";
static int vmstatFlag = 0;

/*******************************************************************************
 * send_dumP_request_to_node()
 *   this basically get the list of all nodes determines which one are data nodes
 *   and send the dump 1000 command to each such that it will generate and event
 *   so the evnet listener will pick it up.
 *
 */
int
send_dump_request_to_nodes(NdbMgmHandle mgmsrv, int node_id ) {
    //static char mod[] = "send_dum_request_to_nodes";
    struct ndb_mgm_reply reply;
    int pars[2] = {1000,0};
    int rtn;

    //printf("%s:-->Called\n", mod);
    /*
     * get the state.
     */
    rtn = ndb_mgm_dump_state(mgmsrv, node_id, pars, 1, &reply);
    //printf("dump sent to %d\n", node_id);
    
    //printf("%s:<--Return\n", mod);
    return 0;

}
/*********************************************************************************
 * printHeader - it it prints the head for what we are displaying
 *
 **/
void 
printHeader(void) {
    if (vmstatFlag) {
        printf("--- NDB ------------------------ ----------Memory ------  -- Swap ---------\n");

        printf("nd typ ps  tpg    pgu  +/-  %%    total    free  cache      free     used\n");
        printf("-------------------------------------------------------------------------------\n");
    } else {
        printf("--------- NDB ---------------------------------------- ----------Trans counts --------------------------------------\n");
        printf("nd typ:ps:totpg:pgu  : %% typ:ps:totpg:pgu  : %%   trans  cmmt  read  sprd  writ  attr  conc  abrt  scan  rang\n");
        printf("--------------------------------------------------------------------------------------------------------------------\n");
        //printf("--- NDB ---------------------- ----------Trans counts --------------------------------------\n");
        //printf("nd typ ps  tpg    pgu +/- %%     trans  cmmt  read  sprd  writ  attr  conc  abrt  scan  rang\n");
        //printf("--------------------------------------------------------------------------------------------\n");
    }
}

/*******************************************************************************
 * send_dumP_request_to_node()
 *   this basically get the list of all nodes determines which one are data nodes
 *   and send the dump 1000 command to each such that it will generate and event
 *   so the evnet listener will pick it up.
 *
 */
int
get_memory_stats(char *bp, int len ) {
    static char mod[] = "get_memory_stats";
    int rtn = 0;
    FILE *f=NULL;
    char buf[80] = {0};
    char *p = buf;
    char *p2;
    unsigned int memTotal = 0;
    unsigned int memFree = 0;
    unsigned int memCached = 0;
    unsigned int swapTotal = 0;
    unsigned int swapFree = 0;
    
    //printf("%s:-->Called\n", mod);

    // open the /proc/memInfo
    if (( f = fopen("/proc/meminfo", "r")) == NULL) {
        printf("ERROR: (%s): %s\n", mod, strerror(errno));
        rtn = -1;
    }
    else {
        while ((p = fgets(buf, sizeof(buf), f)) != NULL) {
            //printf("%s -->%d\n", p, strncmp(p, "MemTotal:", 9));
            // parse the data out of the string
            if (strncmp(p, "MemTotal:", 9) == 0) {
                p2 = strtok(p, " ");
                p2 = strtok(NULL, " ");                memTotal = strtoll(p2, NULL, 0);
            }
            else if (strncmp(p, "MemFree:", 8) == 0) {
                p2 = strtok(p, " ");
                p2 = strtok(NULL, " ");
                memFree = strtoll(p2, NULL, 0);
            }
            else if (strncmp(p, "Cached:", 7) == 0) {
                p2 = strtok(p, " ");
                p2 = strtok(NULL, " ");
                memCached = strtoll(p2, NULL, 0);
            }
            else if (strncmp(p, "SwapFree:", 9) == 0) {
                p2 = strtok(p, " ");
                p2 = strtok(NULL, " ");
                swapFree = strtoll(p2, NULL, 0);
            }
            else if (strncmp(p, "SwapTotal:", 10) == 0) {
                p2 = strtok(p, " ");
                p2 = strtok(NULL, " ");
                swapTotal = strtoll(p2, NULL, 0);
            }
        }
    }
    fclose(f);
    sprintf(bp, "%7u %7u %6u   %6u %6u", memTotal, memFree, memCached, 
            swapFree, (swapTotal - swapFree));
    //printf("%s:<--Returned.\n", mod);

    return rtn;
}
  
/*******************************************************************************
 * usage() - what it is.
 *
 */
void usage(void) {
    printf("usage:%s <-h> -c connectString <-t time_interval_in-seconds>\n", prgName);
}


/*******************************************************************************
 * helpMesg() - Just what is says
 *
 **/
void helpMsg(void) {
    printf("%s version %s\n", prgName, version);
    usage();
    printf("---- NDB --------------------------------------\n");
    printf("nd    - node ID\n");
    printf("typ   - memory type (dm: datameMemory idx: indexMemory\n");
    printf("ps    - page size in kb\n");
    printf("totpg - total pagesn");
    printf("pgu   - pages used\n");
    printf("+/-   - ^ increasing, v descreacing, - no change\n");
    printf("%%     - percent used\n");
    printf("---- Trans counts -----------------------------\n");
    printf("trans - Transaction attempted with this node as coordinator\n");
    printf("cmmt  - Transactions Commited with this node as coordinator\n");
    printf("read  - Primary key reads all\n");
    printf("sprd  - Primary key reads reafing the latest commited value\n");
    printf("writ  - Primary key writes (including all INSERTS,UPDATES, & DELETES)\n");
    printf("attr  - Data words used to describe all reads and writes received\n");
    printf("conc  - All concurrent operations occurring ongoing at the moment the report is taken\n");
    printf("abrt  - Transaction with this node as coordinator that were aborted\n");
    printf("scan  - Scans (all) counts\n");
    printf("rang  - Index scans\n");

    //printf(" -----Memory------------\n");   
    //printf("total - total memory\n");
    //printf("free  - free memory\n");
    //printf("cache - the amount of memory used as cache.\n");
    //printf(" ----- Swap------------\n"); 
    //printf("free  - free swap space\n");
    //printf("used  - swap space in use\n");

    printf("\n");
}
           


/*******************************************************************************
 * main()
 *
 **/
struct option longopts[] =
  {
    { "mgm_host", no_argument, NULL, 'd', },
    { "configFile", required_argument, NULL, 'c', },
    { NULL, },
  };

int main(int argc, char** argv)
{
    int timeInterval = 1;    /* 1 sec */
    NdbMgmHandle h;
    NdbLogEventHandle le;
    char connectstring[60];
    float u, t;
    ndb_mgm_cluster_state *state;
    int i, c;
    unsigned int lineCnt, lastLineCnt;
    char *s;
    int result;
    struct ndb_mgm_reply reply;
    int headerFlag = 0;
    int filter[] = { 15, NDB_MGM_EVENT_CATEGORY_STATISTIC,
		   15, NDB_MGM_EVENT_CATEGORY_CONNECTION,
		   15, NDB_MGM_EVENT_CATEGORY_NODE_RESTART,
		   15, NDB_MGM_EVENT_CATEGORY_STARTUP,
		   15, NDB_MGM_EVENT_CATEGORY_ERROR,
		   0 };
    struct ndb_logevent event;

    /* setup arg0 */
    if (argv[0] == NULL)
        prgName = "mgmvm";
    else {
        if ((s = strrchr(argv[0], '/')) != NULL)
            prgName = ++s;
        else
            prgName = argv[0];
    }

    displayNodeCount = 0;
    
    memset(connectstring, 0, sizeof(connectstring));
    

    while ((c = getopt_long(argc, argv, "hvc:t:", longopts, NULL)) != EOF) {
        switch (c) {
        case 'h':         // -h help
            helpMsg();
            exit(0);
            break;
            
        case 'c':  /* -c connectstring */
            strcpy(connectstring, optarg) ;
            break;

        case 'v': /* -v vmstat */
            vmstatFlag = 1;
            break;
            
        case 't': /* -t time interval */
            timeInterval = atoi(optarg);
            break;

        default:    /* Bad,Bad,Bad arg... */
            usage();
            exit(-1);
            break;
            
        }
    }    

    if (strlen(connectstring) == 0) {
        usage();
        exit(-1);
    }
//     if (argc < 2) {
//         usage();
//         exit(-1);
//     }
//     else if (argc == 3)
//         timeInterval = atoi(argv[1]);

//     connectstring = argv[1];

    printf("%s Version 1.0\n", prgName);
    printf("Starting %s connectstring=%s with %d sec interval time\n", prgName, connectstring, timeInterval);
  
    ndb_init();
  
    //printf("getting the handle\n");
    h= ndb_mgm_create_handle();
    if ( h == 0) {
        printf("Unable to create handle\n");
        exit(-1);
    }
    //printf("setting the connectstring=%s\n", connectstring);
    if (ndb_mgm_set_connectstring(h, connectstring) == -1) {
        printf("Unable to set connectstring\n");
        exit(-1);
    }
    if (ndb_mgm_connect(h,0,0,0)) MGMERROR(h);
    //printf("Connected\n");

    le= ndb_mgm_create_logevent_handle(h, filter);
    if ( le == 0 )  MGMERROR(h);


    lastLineCnt = lineCnt = 0;
    printHeader();

    // zero  out the Nodes area.
    memset(Nodes, 0, sizeof(NODESTATS)*MAX_DISPLAY_NODES);

    while (1) {
        
        state = ndb_mgm_get_status(h);
        if(state == NULL) {
            printf("ERROR: Could not get status\n");
            return(-1);
        }

        //printf("no_of_nodes=%d\n", state->no_of_nodes);
        if ( (lineCnt > (lastLineCnt + LINES_BETWEEN_HEADER))) {

            //if (headerFlag == 0) {
                headerFlag++;
                printHeader();
                lastLineCnt = lineCnt;
                //}
        }
        else 
            headerFlag = 0;
        
        
        for (i = 0 ; i < state->no_of_nodes; i++) {
            // send the "ALL CLUSTERLOG STATISTICS=8" command to all nodes
            if ( (Nodes[i].statCmdSent == 0) && (state->node_states[i].node_type == NDB_MGM_NODE_TYPE_NDB) 
                 && (Nodes[i].dumpSent == 0) ) {            
                result = ndb_mgm_set_loglevel_clusterlog(h,  
                                                         state->node_states[i].node_id, 
                                                         NDB_MGM_EVENT_CATEGORY_STATISTIC,
                                                         8,
                                                         &reply);
                Nodes[i].statCmdSent = 1;
                if (result !=0) {
                    printf("ERROR: sending CLUSTERLOG STATISTIC=8 failed to nodeid=%d.\n",
                           state->node_states[i].node_id );
                    exit(-1);
                }
            }
                                                         

            // printf("%d:nodeID=%d type =%d node_group=%d\n", i,   
//                    state->node_states[i].node_id,
//                    state->node_states[i].node_type,
//                    state->node_states[i].node_group);

            /* 
             * sent the dump cmd to the data node(s) only.
             */
            if ( (state->node_states[i].node_type == NDB_MGM_NODE_TYPE_NDB) && (Nodes[i].dumpSent == 0) ) {            
                send_dump_request_to_nodes(h, state->node_states[i].node_id);
                Nodes[i].dumpSent = 1;
                // make sure that everything is cleared out.
                Nodes[i].dumpDnRecv = Nodes[i].dumpIdxRecv = Nodes[i].statRecv = 0;
            }
        

            while (Nodes[i].dumpSent) {
                int r= ndb_logevent_get_next(le,&event,1000);
                if (r == 0) {
                    //printf("No event within %d seconds\n", timeInterval);
                }
                else if (r < 0)
                    LOGEVENTERROR(le)
                    else {
                        switch (event.type) {
                        case NDB_LE_MemoryUsage:
                            
                            u = (float)event.MemoryUsage.pages_used;
                            t = (float)event.MemoryUsage.pages_total;
                            //printf("u=%f t=%f <%0.2f>\n", u, t, (u/t)*100.0);
                            
                            
                            if (event.MemoryUsage.block == 0xf8) {
                                Nodes[i].dumpIdxRecv = 1;
                                memset(Nodes[i].memidxbuf, 0, 80);
                                sprintf(Nodes[i].memidxbuf, "%s:%02d:%05u:%05u:%2.0f %c",
                                        (event.MemoryUsage.block==0xf8 ? "idx" : (event.MemoryUsage.block == 0xf9 ?" dm":"unk")),
                                        event.MemoryUsage.page_size_kb/1000,
                                        event.MemoryUsage.pages_total,
                                        event.MemoryUsage.pages_used,
                                        (u/t)*100.0,
                                        (event.MemoryUsage.gth == 0 ? '-': (event.MemoryUsage.gth > 0 ? '^':'v'))
                                        );
//                                 sprintf(Nodes[i].memidxbuf, "%2d %s %2d %5u %5u  %c  %2.0f  ",
//                                         event.source_nodeid,
//                                         (event.MemoryUsage.block==0xf8 ? "idx" : (event.MemoryUsage.block == 0xf9 ?" dm":"unk")),
//                                         event.MemoryUsage.page_size_kb/1000,
//                                         event.MemoryUsage.pages_total,
//                                         event.MemoryUsage.pages_used,
//                                         (event.MemoryUsage.gth == 0 ? '-': (event.MemoryUsage.gth > 0 ? '^':'v')),
//                                         (u/t)*100.0);
                            }
                            else {
                                Nodes[i].dumpDnRecv = 1;
                                memset(Nodes[i].memdatabuf, 0, 80);
                                sprintf(Nodes[i].memdatabuf, "%2d:%s:%2d:%05u:%05u:%2.0f",
                                        event.source_nodeid,
                                        (event.MemoryUsage.block==0xf8 ? "idx" : (event.MemoryUsage.block == 0xf9 ?" dm":"unk")),
                                        event.MemoryUsage.page_size_kb/1000,
                                        event.MemoryUsage.pages_total,
                                        event.MemoryUsage.pages_used,
                                        (u/t)*100.0);
//                                 sprintf(Nodes[i].memdatabuf, "%2d %s %2d %5u %5u  %c  %2.0f  ",
//                                         event.source_nodeid,
//                                         (event.MemoryUsage.block==0xf8 ? "idx" : (event.MemoryUsage.block == 0xf9 ?" dm":"unk")),
//                                         event.MemoryUsage.page_size_kb/1000,
//                                         event.MemoryUsage.pages_total,
//                                         event.MemoryUsage.pages_used,
//                                         (event.MemoryUsage.gth == 0 ? '-': (event.MemoryUsage.gth > 0 ? '^':'v')),
//                                         (u/t)*100.0);
                            }
                            
                            if (vmstatFlag) {
                                //len = strlen(p->membuf);
                                //get_memory_stats(&p->membuf[len], sizeof(80) - len);
                            }
                            //printf("%s\n", p->membuf);
                            break;
                            
                        case NDB_LE_TransReportCounters:
                            if (vmstatFlag == 0) {
                                Nodes[i].statRecv = 1;
                                
                                memset(Nodes[i].countbuf, 0, 80);
                                
                                sprintf(Nodes[i].countbuf, "%05u %05u %05u %05u %05u %05u %05u %05u %05u %05u", 
                                        event.TransReportCounters.trans_count, 
                                        event.TransReportCounters.commit_count, 
                                        event.TransReportCounters.read_count, 
                                        event.TransReportCounters.simple_read_count,
                                        event.TransReportCounters.write_count,
                                        event.TransReportCounters.attrinfo_count,
                                        event.TransReportCounters.conc_op_count,
                                        event.TransReportCounters.abort_count, 
                                        event.TransReportCounters.scan_count, 
                                        event.TransReportCounters.range_scan_count);
                            }
                            //printf("ndoeid=%d %s\n", event.source_nodeid, p->countbuf);
                            break;
                            
                        case NDB_LE_OperationReportCounters:
                        case NDB_LE_ReceiveBytesStatistic:
                        case NDB_LE_SendBytesStatistic:
                            //printf("TEST:nodeid=%d\n", event.source_nodeid);
                            break;
                        case NDB_LE_BackupStarted:
                            printf("Node %d: BackupStarted\n", event.source_nodeid);
                            printf("  Starting node ID: %d\n", event.BackupStarted.starting_node);
                            printf("  Backup ID: %d\n", event.BackupStarted.backup_id);
                            break;
                        case NDB_LE_BackupCompleted:
                            printf("Node %d: BackupCompleted\n", event.source_nodeid);
                            printf("  Backup ID: %d\n", event.BackupStarted.backup_id);
                            break;
                        case NDB_LE_BackupAborted:
                            printf("Node %d: BackupAborted\n", event.source_nodeid);
                            break;
                        case NDB_LE_BackupFailedToStart:
                            printf("Node %d: BackupFailedToStart\n", event.source_nodeid);
                            break;
                            
                        case NDB_LE_NodeFailCompleted:
                            printf("Node %d: NodeFailCompleted\n", event.source_nodeid);
                            break;
                        case NDB_LE_ArbitResult:
                            printf("Node %d: ArbitResult\n", event.source_nodeid);
                            printf("  code %d, arbit_node %d\n",
                                   event.ArbitResult.code & 0xffff,
                                   event.ArbitResult.arbit_node);
                            break;
                        case NDB_LE_DeadDueToHeartbeat:
                            printf("Node %d: DeadDueToHeartbeat\n", event.source_nodeid);
                            printf("  node %d\n", event.DeadDueToHeartbeat.node);
                            break;
                            
                        case NDB_LE_Connected:
                            printf("Node %d: Connected\n", event.source_nodeid);
                            printf("  node %d\n", event.Connected.node);
                            Nodes[i].nodeState = 1;
                            break;
                        case NDB_LE_Disconnected:
                            printf("Node %d: Disconnected\n", event.source_nodeid);
                            printf("  node %d\n", event.Disconnected.node);
                            Nodes[i].nodeState = 0;
                            break;
                        case NDB_LE_NDBStartCompleted:
                            printf("Node %d: StartCompleted\n", event.source_nodeid);
                            printf("  version %d.%d.%d\n",
                                   event.NDBStartCompleted.version >> 16 & 0xff,
                                   event.NDBStartCompleted.version >> 8 & 0xff,
                                   event.NDBStartCompleted.version >> 0 & 0xff);
                            break;
                        case NDB_LE_ArbitState:
                            printf("Node %d: ArbitState\n", event.source_nodeid);
                            printf("  code %d, arbit_node %d\n",
                                   event.ArbitState.code & 0xffff,
                                   event.ArbitResult.arbit_node);
                            break;
                            
                        default:
                            break;
                        }
                    }
                // make sure we have all the data and print.
                if ( (Nodes[i].dumpDnRecv && Nodes[i].statRecv) 
                     && (Nodes[i].dumpIdxRecv && Nodes[i].statRecv) ) {
                      printf("%s %s %s \n", Nodes[i].memdatabuf, Nodes[i].memidxbuf, Nodes[i].countbuf);
                      lineCnt++;
                }
//                 if (Nodes[i].dumpIdxRecv && Nodes[i].statRecv) {
//                     printf("%s %s %s\n", Nodes[i].memidxbuf, Nodes[i].countbuf);
//                     lineCnt++;
//                 }
                
                if (Nodes[i].dumpDnRecv && Nodes[i].statRecv && Nodes[i].dumpIdxRecv && Nodes[i].dumpSent)
                    Nodes[i].dumpSent = 0;
            } // while ..........
                

            //printf("headerFlag = %d linecnt=%d last=%d\n", headerFlag, lineCnt, lastLineCnt);
            //if (printLine())
            //    lineCnt++;
        }
        sleep(timeInterval);
    }
      
    ndb_mgm_destroy_logevent_handle(&le);
    ndb_mgm_destroy_handle(&h);
    ndb_end(0);
    return 0;
}
